# -*- coding: utf-8 -*-

# Copyright 2021 Jean-Baptiste Delisle
#
# This file is part of samsam.
#
# samsam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# samsam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with samsam.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.special import beta as sp_beta, erf as sp_erf

log2pi = np.log(2.0 * np.pi)
halflog2pi = log2pi / 2.0


def uniform(x, a=0.0, b=1.0):
  r"""
  Uniform distribution over an interval :math:`[a,b[`.

  Parameters
  ----------
  x : float
    Value of the variable.
  a : float
    Interval lower bound.
  b : float
    Interval upper bound.

  Returns
  -------
  lp : float
    Log-probability of x.
  """

  if x < a or x >= b:
    raise Exception('uniform: out of bounds.')
  return (-np.log(b - a))


def uniform_periodic(x, period):
  r"""
  Uniform distribution over a periodic interval.

  Parameters
  ----------
  x : float
    Value of the variable.
  period : float
    Periodicity.

  Returns
  -------
  lp : float
    Log-probability of x.
  """

  return (-np.log(period))


def loguniform(x, loga=0.0, logb=1.0):
  r"""
  Log-uniform distribution over an interval :math:`[a,b[`.

  Parameters
  ----------
  x : float
    Value of the variable.
  loga : float
    Logarithm of the lower bound.
  logb : float
    Logarithm of the upper bound.

  Returns
  -------
  lp : float
    Log-probability of x.
  """

  if x <= 0:
    raise Exception('loguniform: out of bounds.')
  logx = np.log(x)
  return (-logx + uniform(logx, loga, logb))


def moduniform(x, y, a=0.0, b=1.0):
  r"""
  Uniform distribution for the module :math:`\sqrt{x^2+y^2}`
  over an interval :math:`[a,b[`.

  Parameters
  ----------
  x : float
    Value of the abscissa.
  y : float
    Value of the ordinate.
  a : float
    Interval lower bound.
  b : float
    Interval upper bound.

  Returns
  -------
  lp : float
    Log-probability of x.
  """

  r = np.sqrt(x**2 + y**2)
  if r <= a or r >= b:
    raise Exception('moduniform: out of bounds.')
  return (-(log2pi + np.log(r * (b - a))))


def normal(x, mu=0.0, sig=1.0):
  r"""
  Normal distribution with parameters (mu,sig).

  Parameters
  ----------
  x : float
    Value of the variable.
  mu : float
    Mean.
  sig : float
    Standard-deviation.

  Returns
  -------
  lp : float
    Log-probability of x.
  """

  return (-(halflog2pi + np.log(sig) + ((x - mu) / sig)**2 / 2.0))


logtruncnorm_dic = {}


def _lazy_logtruncnorm(mu, sig, a, b):
  r"""
  Lazy computation of the normalizing coefficient for truncnormal.
  """
  if (mu, sig, a, b) not in logtruncnorm_dic:
    xa = (a - mu) / (np.sqrt(2) * sig)
    xb = (b - mu) / (np.sqrt(2) * sig)
    logtruncnorm_dic[(mu, sig, a,
      b)] = -np.log(np.sqrt(np.pi / 2) * sig * (sp_erf(xb) - sp_erf(xa)))
  return (logtruncnorm_dic[(mu, sig, a, b)])


def truncnormal(x, mu=0.0, sig=1.0, a=0, b=np.inf):
  r"""
  Truncated normal distribution with parameters (mu,sig,a,b).

  Parameters
  ----------
  x : float
    Value of the variable.
  mu : float
    Mean.
  sig : float
    Standard-deviation.
  a : float
    Lower bound.
  b : float
    Upper bound.

  Returns
  -------
  lp : float
    Log-probability of x.
  """

  if x < a or x >= b:
    raise Exception('truncnormal: out of bounds.')
  return (_lazy_logtruncnorm(mu, sig, a, b) - ((x - mu) / sig)**2 / 2.0)


def lognormal(x, mu=0.0, sig=1.0):
  r"""
  Log-normal distribution with parameters (mu,sig).

  Parameters
  ----------
  x : float
    Value of the variable.
  mu : float
    Mean of :math:`\log(x)`.
  sig : float
    Standard-deviation of :math:`\log(x)`.

  Returns
  -------
  lp : float
    Log-probability of x.
  """

  if x <= 0:
    raise Exception('lognormal: out of bounds.')
  logx = np.log(x)
  return (-logx + normal(logx, mu, sig))


logbeta_dic = {}


def _lazy_logbeta(a, b):
  r"""
  Lazy computation of log(beta(a,b))
  """
  if (a, b) not in logbeta_dic:
    logbeta_dic[(a, b)] = np.log(sp_beta(a, b))
  return (logbeta_dic[(a, b)])


def beta(x, a, b):
  r"""
  Beta distribution with parameters (a, b).

  Parameters
  ----------
  x : float
    Value of the variable.
  a, b : float
    Shape parameters.

  Returns
  -------
  lp : float
    Log-probability of x.
  """

  if x <= 0 or x >= 1:
    raise Exception('beta: out of bounds.')
  return ((a - 1.0) * np.log(x) + (b - 1.0) * np.log(1.0 - x) -
    _lazy_logbeta(a, b))


def modbeta(x, y, a, b):
  r"""
  Beta distribution with parameters (a, b)
  for the module :math:`\sqrt{x^2+y^2}`.

  Parameters
  ----------
  x : float
    Value of the abscissa.
  y : float
    Value of the ordinate.
  a, b : float
    Shape parameters.

  Returns
  -------
  lp : float
    Log-probability of x.
  """

  r = np.sqrt(x**2 + y**2)
  if r == 0 or r >= 1:
    raise Exception('modbeta: out of bounds.')
  return ((a - 2.0) * np.log(r) + (b - 1.0) * np.log(1.0 - r) -
    _lazy_logbeta(a, b) - log2pi)



# -*- coding: utf-8 -*-

# Copyright 2021 Jean-Baptiste Delisle
#
# This file is part of samsam.
#
# samsam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# samsam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with samsam.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np


def acf(x):
  r"""
  Auto-correlation function of the time series x.

  Parameters
  ----------
  x : (n,) or (n, d) ndarray
    Time series values (which can be multi-dimensional).

  Returns
  -------
  R : (n,) or (n, d) ndarray
    Auto-correlation as a function of lag.
  """

  nt = x.shape[0]
  f = np.fft.fft(x - np.mean(x, axis=0), axis=0, n=2 * nt)
  S = abs(f)**2
  R = np.fft.ifft(S, axis=0)[:nt].real
  R /= R[0]
  return (R)


def iat(x=None, R=None, window=50):
  r"""
  Integrated auto-correlation time of the time series x.

  Parameters
  ----------
  x : (n,) or (n, d) ndarray or None
    Time series values (which can be multi-dimensional).
    If R = acf(x) is provided, this is not used.
  R : (n,) or (n, d) ndarray or None
    Auto-correlation as a function of lag.
  window : int
    Window used to smooth the acf.

  Returns
  -------
  IAT : float or (d,) ndarray
    Integrated auto-correlation time (for each dimension).
  """

  if R is None:
    R = acf(x)
  nt = len(R)
  nwindow = nt // window
  nt = nwindow * window
  if R.ndim == 1:
    Ga = np.sum(R[:nt].reshape(nwindow, window), axis=1)
    dGa = Ga[1:] - Ga[:-1]
    kneg = np.where(Ga <= 0)[0][0]
    kincr = np.where(dGa >= 0)[0][0]
    kmax = min(kneg, kincr)
    IAT = 2 * np.sum(Ga[:kmax]) - 1
  else:
    nDim = R.shape[1]
    Ga = np.sum(R[:nt].reshape(nwindow, window, nDim), axis=1)
    dGa = Ga[1:] - Ga[:-1]
    IAT = np.empty(nDim)
    for c in range(nDim):
      kneg = np.where(Ga[:, c] <= 0)[0][0]
      kincr = np.where(dGa[:, c] >= 0)[0][0]
      kmax = min(kneg, kincr)
      IAT[c] = 2 * np.sum(Ga[:kmax, c]) - 1
  return (IAT)



# -*- coding: utf-8 -*-

# Copyright 2021 Jean-Baptiste Delisle
#
# This file is part of samsam.
#
# samsam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# samsam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with samsam.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from .__info__ import __version__


def sam(x0,
  logprob,
  nsamples=100000,
  mu0=None,
  cov0=None,
  cov_update_interval=100,
  cov_update_law=lambda t: (100 / (t + 1))**(2 / 3),
  scale0=None,
  accept_rate_target=0.234,
  force_use_empirical_cov=False,
  print_level=1,
  print_interval=1000,
  print_inplace=True,
  **kwargs):
  r"""
  Adaptive Metropolis algorithm (Haario et al. 2001)
  with adaptive scaling (Andrieu & Thoms 2008).

  Parameters
  ----------
  x0 : (ndim,) ndarray
    Initial guess of the parameters.
  logprob(x, **kwargs) : function
    Log probability of the distribution to sample.
  nsamples: int
    Number of samples to draw.
  mu0 : (ndim,) ndarray
    Initial guess of the parameters' mean.
  cov0 : (ndim, ndim) ndarray
    Initial guess of the parameters' covariance matrix.
  cov_update_interval : int
    Interval at which to update the covariance matrix.
  cov_update_law(t) : function
    Update coefficient (between 0 and 1)
    as a function of time (should vanish as t -> nsamples)
  scale0 : float
    Initial scaling factor for the proposal distribution
    (by default 2.4/sqrt(ndim)).
  accept_rate_target : float
    Acceptance rate to target when rescaling the proposal distribution,
    if None, the scale is left unchanged.
  force_use_empirical_cov : bool
    Whether to force the use of the empirical covariance from the beginning.
    If this is not forced, we wait for the acceptance rate to reach at least
    0.1 before considering the empirical covariance to be correct.
  print_level : int
    0 (no printing)
    1 (print acceptance rate)
  print_interval : int
    Interval at which to print infos.
  print_inplace : bool
    Whether to print infos in place or one line after the other.
  **kwargs :
    Additional parameters for the logprob function.

  Returns
  -------
  samples : (nsamples+1, ndim) ndarray
    Array of parameters values for each sample.
  diagnositics : dict
    Dictionary of diagnostics, with the following keys:

      logprob : (nsamples+1,) ndarray
        Array of log probability for each sample.
      alpha : (nsamples,) ndarray
        Array of acceptance probability for each proposal.
      accept : (nsamples,) ndarray
        Array of acceptance for each proposal.
      mu : (ndim,) ndarray
        Final estimate of the mean.
      cov : (ndim, ndim) ndarray
        Final estimate of the covariance matrix.
      scale : float
        Final estimate of the proposal scale.
  """

  nsamples = int(nsamples)
  ndim = len(x0)
  # Init state
  x = x0
  lpx = logprob(x, **kwargs)
  # Init chain
  histx = np.empty((nsamples + 1, ndim))
  histlpx = np.empty(nsamples + 1)
  histalpha = np.empty(nsamples)
  histaccept = np.empty(nsamples, dtype=bool)
  histx[0] = x
  histlpx[0] = lpx
  # Init covariance matrix
  if mu0 is None:
    mu = x0.copy()
  else:
    mu = mu0.copy()
  if cov0 is None:
    C = np.identity(ndim)
  else:
    if cov0.shape != (ndim, ndim):
      raise Exception(
        'Incompatible shapes for x0 ({:d}) and cov0 {:s}.'.format(
        ndim, str(cov0.shape)))
    C = cov0
  if scale0 is None:
    scale = 2.4 / np.sqrt(ndim)
  else:
    scale = scale0
  # SVD decomposition of C (more robust than Cholesky)
  _, s, v = np.linalg.svd(C)
  sqCT = np.sqrt(s)[:, None] * v
  use_empirical_cov = force_use_empirical_cov
  # Init printing
  print_fmt = (
    '{}Step {{:{:d}d}}, acceptance rate (since last printing): {{:.4f}}'.
    format('\r' if print_inplace else '', 1 + int(np.log10(nsamples))))
  print_end = ' ' if print_inplace else None

  # Big loop
  for t in range(1, nsamples + 1):
    # Proposal of new point (y)
    y = x + np.random.normal(scale=scale, size=ndim).dot(sqCT)
    # Compute proposal probability
    lpy = logprob(y, **kwargs)
    # Do we accept the proposal
    alpha = np.exp(min(0.0, lpy - lpx))
    accept = np.random.random() < alpha
    if accept:
      x = y
      lpx = lpy
    # Save state in chain
    histx[t] = x
    histlpx[t] = lpx
    histalpha[t - 1] = alpha
    histaccept[t - 1] = accept
    # Update covariance matrix
    if t % cov_update_interval == 0:
      gamma = cov_update_law(t)
      # Update mean
      mudt = np.mean(histx[t + 1 - cov_update_interval:t + 1], axis=0)
      dmu = mudt - mu
      mu += gamma * dmu
      # Update cov
      Cdt = np.cov(histx[t + 1 - cov_update_interval:t + 1], rowvar=False)
      dmu.shape = (ndim, 1)
      C = (1.0 - gamma) * C + gamma * Cdt + gamma * (1.0 - gamma) * dmu.dot(
        dmu.T)
      if not use_empirical_cov:
        mean_accept = np.mean(histaccept[t - cov_update_interval:t])
        if mean_accept > 0.1:
          use_empirical_cov = True
        else:
          coef = (mean_accept + 0.025) / 0.125
          C *= coef * coef
          sqCT *= coef
      if use_empirical_cov:
        # SVD decomposition of C (more robust than Cholesky)
        _, s, v = np.linalg.svd(C)
        sqCT = np.sqrt(s)[:, None] * v
      # Adapt scale
      if accept_rate_target is not None and use_empirical_cov:
        mean_alpha = np.mean(histalpha[t - cov_update_interval:t])
        scale *= ((mean_alpha + 0.25 * accept_rate_target) /
          (1.25 * accept_rate_target))**gamma
    # Print infos
    if print_level and t % print_interval == 0:
      print(print_fmt.format(t, np.mean(histaccept[t - print_interval:t])),
        end=print_end)
  if print_level and print_inplace and nsamples >= print_interval:
    print()
  return (histx, {
    'logprob': histlpx,
    'alpha': histalpha,
    'accept': histaccept,
    'mu': mu,
    'cov': C,
    'scale': scale
  })


def covis(mu,
  cov,
  logprob,
  nsamples=100000,
  print_level=1,
  print_interval=1000,
  print_inplace=True,
  **kwargs):
  """
  Importance sampling using a multivariate normal sampling distribution.

  The covis sampler generates samples from a normal distribution
  with mean `mu` and covariance `cov`.
  For each sample, the target distribution `logprob` is evaluated
  and a weight is deduced which allows to estimate integrals
  over the target distribution (e.g. the evidence).
  The parameters `mu` and `cov` should be chosen such that the sampling
  distribution is close to the target distribution.

  Parameters
  ----------
  mu : (ndim,) ndarray
    Mean of the sampling distribution.
  cov :  (ndim, ndim) ndarray
    Covariacne of the sampling distribution
  logprob(x, **kwargs) : function
    Log probability of the distribution to sample.
  nsamples: int
    Number of samples to draw.
  print_level : int
    0 (no printing)
    1 (print step)
  print_interval : int
    Interval at which to print infos.
  print_inplace : bool
    Whether to print infos in place or one line after the other.
  **kwargs :
    Additional parameters for the logprob function.

  Returns
  -------
  samples : (nsamples, ndim) ndarray
    Array of parameters values for each sample.
  logweights : (nsamples,) ndarray
    Array of samples log weight.
  diagnositics : dict
    Dictionary of diagnostics, with the following keys:

      logsamp : (nsamples,) ndarray
        Array of log probability of each sample for the sampling distribution.
      logprob : (nsamples,) ndarray
        Array of log probability of each sample for the target distribution.
      logevidence : float
        Log evidence of the target distribution.
  """

  ndim = mu.size
  if cov.shape != (ndim, ndim):
    raise Exception('Incompatible shapes for mu ({:d}) and cov {:s}.'.format(
      ndim, str(cov.shape)))

  # SVD decomposition of cov (more robust than Cholesky)
  _, s, v = np.linalg.svd(cov)
  sqCT = np.sqrt(s)[:, None] * v
  logdet2piC = np.sum(np.log(2 * np.pi * s))

  # Init arrays
  histx = np.empty((nsamples, ndim))
  histlsampx = np.empty(nsamples)
  histlpx = np.empty(nsamples)
  histlw = np.empty(nsamples)

  # Init printing
  print_fmt = ('{}Step {{:{:d}d}}'.format('\r' if print_inplace else '',
    1 + int(np.log10(nsamples))))
  print_end = ' ' if print_inplace else None

  # Big loop
  for k in range(nsamples):
    # Generate sample
    ux = np.random.normal(size=ndim)
    x = mu + ux.dot(sqCT)
    # Compute sampling probabilty
    histlsampx[k] = -(logdet2piC + np.sum(ux * ux)) / 2
    # Compute target probability
    histlpx[k] = logprob(x, **kwargs)
    # Log weight
    histlw[k] = histlpx[k] - histlsampx[k]
    if print_level and (k + 1) % print_interval == 0:
      print(print_fmt.format(k + 1), end=print_end)
  if print_level and print_inplace and nsamples >= print_interval:
    print()

  # Compute log evidence
  # Find max log weight to avoid under/overflows
  maxlw = np.max(histlw)
  lZ = maxlw + np.log(np.sum(np.exp(histlw - maxlw)) / nsamples)

  return (histx, histlw, {
    'logsamp': histlsampx,
    'logprob': histlpx,
    'logevidence': lZ
  })



import pytest
import numpy as np
from samsam import sam, covis, logprior, acf

n = 4
nsamples = 100000


def generate_mucov(seed=0):
  np.random.seed(seed)
  mu = np.random.normal(0, 10, n)
  cov = np.identity(n) + np.random.normal(0, 1e-2, n**2).reshape((n, n))
  cov = (cov + cov.T) / 2
  return (mu, cov)


def logprob(x, x_mu, x_cov):
  r = x - x_mu
  return (-0.5 *
    (r @ np.linalg.inv(x_cov) @ r + np.linalg.slogdet(2 * np.pi * x_cov)[1]))


def test_sam():
  mu, cov = generate_mucov()
  x0 = np.zeros(n)
  samples, _ = sam(x0, logprob, nsamples=nsamples, x_mu=mu, x_cov=cov)
  samples = samples[nsamples // 4::10]
  samp_mu = np.mean(samples, axis=0)
  samp_cov = np.cov(samples, rowvar=False)

  assert np.max(np.abs(samp_mu - mu)) < 0.1
  assert np.max(np.abs(samp_cov - cov)) < 0.1


def test_covis():
  mu, cov = generate_mucov()
  mu0 = mu + np.random.normal(0, 0.1, n)
  cov0 = 0.9 * cov + np.random.normal(0, 1e-3, n**2).reshape((n, n))
  cov0 = (cov0 + cov0.T) / 2

  _, _, diags = covis(mu0,
    cov0,
    logprob,
    nsamples=nsamples,
    x_mu=mu,
    x_cov=cov)

  assert abs(diags['logevidence']) < 1e-3
